#ifndef STAN_MATH_HPP
#define STAN_MATH_HPP

/**
 * \defgroup prob_dists Probability Distributions
 */

/**
 * \ingroup prob_dists
 * \defgroup multivar_dists Multivariate Distributions
 * Distributions with Matrix inputs
 */
/**
 * \ingroup prob_dists
 * \defgroup univar_dists Univariate Distributions
 * Distributions with scalar, vector, or array input.
 */

#include <stan/math/rev.hpp>

#endif
