#ifndef STAN_MATH_OPENCL_ERR_HPP
#define STAN_MATH_OPENCL_ERR_HPP

#include <stan/math/opencl/err/check_mat_size_one.hpp>
#include <stan/math/opencl/err/check_opencl.hpp>
#include <stan/math/opencl/err/check_symmetric.hpp>
#include <stan/math/opencl/err/check_triangular.hpp>

#endif
