#ifndef STAN_MATH_MIX_FUNCTOR_HPP
#define STAN_MATH_MIX_FUNCTOR_HPP

#include <stan/math/mix/functor/derivative.hpp>
#include <stan/math/mix/functor/finite_diff_grad_hessian.hpp>
#include <stan/math/mix/functor/finite_diff_grad_hessian_auto.hpp>
#include <stan/math/mix/functor/grad_hessian.hpp>
#include <stan/math/mix/functor/grad_tr_mat_times_hessian.hpp>
#include <stan/math/mix/functor/gradient_dot_vector.hpp>
#include <stan/math/mix/functor/hessian.hpp>
#include <stan/math/mix/functor/hessian_times_vector.hpp>
#include <stan/math/mix/functor/partial_derivative.hpp>

#endif
