#ifndef STAN_MATH_PRIM_ERR_HPP
#define STAN_MATH_PRIM_ERR_HPP

#include <stan/math/prim/err/check_2F1_converges.hpp>
#include <stan/math/prim/err/check_3F2_converges.hpp>
#include <stan/math/prim/err/check_bounded.hpp>
#include <stan/math/prim/err/check_cholesky_factor.hpp>
#include <stan/math/prim/err/check_cholesky_factor_corr.hpp>
#include <stan/math/prim/err/check_column_index.hpp>
#include <stan/math/prim/err/check_consistent_size.hpp>
#include <stan/math/prim/err/check_consistent_sizes.hpp>
#include <stan/math/prim/err/check_consistent_sizes_mvt.hpp>
#include <stan/math/prim/err/check_corr_matrix.hpp>
#include <stan/math/prim/err/check_cov_matrix.hpp>
#include <stan/math/prim/err/check_finite.hpp>
#include <stan/math/prim/err/check_flag_sundials.hpp>
#include <stan/math/prim/err/check_greater.hpp>
#include <stan/math/prim/err/check_greater_or_equal.hpp>
#include <stan/math/prim/err/check_ldlt_factor.hpp>
#include <stan/math/prim/err/check_less.hpp>
#include <stan/math/prim/err/check_less_or_equal.hpp>
#include <stan/math/prim/err/check_lower_triangular.hpp>
#include <stan/math/prim/err/check_matching_dims.hpp>
#include <stan/math/prim/err/check_matching_sizes.hpp>
#include <stan/math/prim/err/check_multiplicable.hpp>
#include <stan/math/prim/err/check_nonnegative.hpp>
#include <stan/math/prim/err/check_nonzero_size.hpp>
#include <stan/math/prim/err/check_not_nan.hpp>
#include <stan/math/prim/err/check_ordered.hpp>
#include <stan/math/prim/err/check_sorted.hpp>
#include <stan/math/prim/err/check_pos_definite.hpp>
#include <stan/math/prim/err/check_pos_semidefinite.hpp>
#include <stan/math/prim/err/check_positive.hpp>
#include <stan/math/prim/err/check_positive_finite.hpp>
#include <stan/math/prim/err/check_positive_ordered.hpp>
#include <stan/math/prim/err/check_range.hpp>
#include <stan/math/prim/err/check_row_index.hpp>
#include <stan/math/prim/err/check_simplex.hpp>
#include <stan/math/prim/err/check_size_match.hpp>
#include <stan/math/prim/err/check_square.hpp>
#include <stan/math/prim/err/check_std_vector_index.hpp>
#include <stan/math/prim/err/check_symmetric.hpp>
#include <stan/math/prim/err/check_unit_vector.hpp>
#include <stan/math/prim/err/check_vector.hpp>
#include <stan/math/prim/err/check_vector_index.hpp>
#include <stan/math/prim/err/constraint_tolerance.hpp>
#include <stan/math/prim/err/domain_error.hpp>
#include <stan/math/prim/err/domain_error_vec.hpp>
#include <stan/math/prim/err/elementwise_check.hpp>
#include <stan/math/prim/err/hmm_check.hpp>
#include <stan/math/prim/err/invalid_argument.hpp>
#include <stan/math/prim/err/invalid_argument_vec.hpp>
#include <stan/math/prim/err/is_cholesky_factor.hpp>
#include <stan/math/prim/err/is_cholesky_factor_corr.hpp>
#include <stan/math/prim/err/is_column_index.hpp>
#include <stan/math/prim/err/is_corr_matrix.hpp>
#include <stan/math/prim/err/is_ldlt_factor.hpp>
#include <stan/math/prim/err/is_less_or_equal.hpp>
#include <stan/math/prim/err/is_lower_triangular.hpp>
#include <stan/math/prim/err/is_mat_finite.hpp>
#include <stan/math/prim/err/is_matching_dims.hpp>
#include <stan/math/prim/err/is_matching_size.hpp>
#include <stan/math/prim/err/is_nonzero_size.hpp>
#include <stan/math/prim/err/is_not_nan.hpp>
#include <stan/math/prim/err/is_ordered.hpp>
#include <stan/math/prim/err/is_pos_definite.hpp>
#include <stan/math/prim/err/is_positive.hpp>
#include <stan/math/prim/err/is_scal_finite.hpp>
#include <stan/math/prim/err/is_size_match.hpp>
#include <stan/math/prim/err/is_square.hpp>
#include <stan/math/prim/err/is_symmetric.hpp>
#include <stan/math/prim/err/is_unit_vector.hpp>
#include <stan/math/prim/err/out_of_range.hpp>
#include <stan/math/prim/err/system_error.hpp>
#include <stan/math/prim/err/throw_domain_error.hpp>
#include <stan/math/prim/err/throw_domain_error_mat.hpp>
#include <stan/math/prim/err/throw_domain_error_vec.hpp>
#include <stan/math/prim/err/validate_non_negative_index.hpp>
#include <stan/math/prim/err/validate_positive_index.hpp>
#include <stan/math/prim/err/validate_unit_vector_index.hpp>
#endif
