#ifndef STAN_MATH_PRIM_HPP
#define STAN_MATH_PRIM_HPP

#include <stan/math/version.hpp>

#ifdef STAN_OPENCL
#include <stan/math/opencl/prim.hpp>
#endif

#include <stan/math/prim/core.hpp>
#include <stan/math/prim/meta.hpp>
#include <stan/math/prim/err.hpp>

#include <stan/math/prim/fun.hpp>
#include <stan/math/prim/functor.hpp>
#include <stan/math/prim/prob.hpp>

#endif
